package morechart.chart.renderer.xy

import java.awt. { BasicStroke, Graphics2D, Image, Toolkit }
import java.awt.image.BufferedImage
import java.awt.geom.Rectangle2D

import morechart.chart.renderer. { LookupSVGScale, SVGScale }

import org.apache.batik.dom.svg.SVGDOMImplementation

import org.apache.batik.transcoder. {

  SVGAbstractTranscoder,
  TranscodingHints,
  TranscoderInput,
  TranscoderOutput,
  XMLAbstractTranscoder

}

import org.apache.batik.transcoder.image.ImageTranscoder
import org.apache.batik.util.SVGConstants

import org.jfree.data.xy. { XYDataset, XYZDataset }
import org.jfree.chart.axis.ValueAxis

import org.jfree.chart.plot. {

  CrosshairState,
  PlotOrientation,
  PlotRenderingInfo,
  XYPlot

}

import org.jfree.chart.renderer.xy. { XYBlockRenderer, XYItemRendererState }

import scala.collection.mutable.Map

/** Extends [[org.jfree.chart.renderer.xy.XYBlockRenderer]] with svg.
 *
 * Use cache method to prevent resizing of same item images.
 *
 * @author myst3r10n
 */
class XYSVGRenderer extends XYBlockRenderer {

  protected var _svgScale = new LookupSVGScale

  /** @see [[org.jfree.chart.renderer.xy.XYBlockRenderer]]. */
  def svgScale = _svgScale

  /** @see [[org.jfree.chart.renderer.xy.XYBlockRenderer]]. */
  def svgScale_=(scale: SVGScale) {

    if(!cachedItems.isEmpty)
      cachedItems.clear

    _svgScale = scale.asInstanceOf[LookupSVGScale]

  }

  /** @see [[org.jfree.chart.renderer.xy.XYBlockRenderer]]. */
  override def drawItem(g2: Graphics2D,
                          state: XYItemRendererState,
                          dataArea: Rectangle2D,
                          info: PlotRenderingInfo,
                          plot: XYPlot,
                          domainAxis: ValueAxis,
                          rangeAxis: ValueAxis,
                          dataset: XYDataset,
                          series: Int,
                          item: Int,
                          crosshairState: CrosshairState,
                          pass: Int) {

    // XYBlockRenderer part.
    val x = dataset.getXValue(series, item)
    val y = dataset.getYValue(series, item)
    var z = 0.0;
    if(dataset.isInstanceOf[XYZDataset])
      z = dataset.asInstanceOf[XYZDataset].getZValue(series, item)

    val xOffset = classOf[XYBlockRenderer].getDeclaredField("xOffset")
    val yOffset = classOf[XYBlockRenderer].getDeclaredField("yOffset")
    xOffset.setAccessible(true)
    yOffset.setAccessible(true)

    val xx0 = domainAxis.valueToJava2D(x + xOffset.get(this).asInstanceOf[Double],
                                       dataArea, plot.getDomainAxisEdge)
    val yy0 = rangeAxis.valueToJava2D(y + yOffset.get(this).asInstanceOf[Double],
                                      dataArea, plot.getRangeAxisEdge)
    val xx1 = domainAxis.valueToJava2D(x + getBlockWidth + xOffset.get(this).asInstanceOf[Double],
                                       dataArea, plot.getDomainAxisEdge)
    val yy1 = rangeAxis.valueToJava2D(y + getBlockHeight + yOffset.get(this).asInstanceOf[Double],
                                      dataArea, plot.getRangeAxisEdge)
    var block: Rectangle2D = null
    if(plot.getOrientation.equals(PlotOrientation.HORIZONTAL))
      block = new Rectangle2D.Double(math.min(yy0, yy1),
                                     math.min(xx0, xx1),
                                     math.abs(yy1 - yy0),
                                     math.abs(xx0 - xx1))
    else
      block = new Rectangle2D.Double(math.min(xx0, xx1),
                                     math.min(yy0, yy1),
                                     math.abs(xx1 - xx0),
                                     math.abs(yy1 - yy0))
    g2.setStroke(new BasicStroke(1.0f))

    // Draw SVG.
    val size = new Dimension2D.Float(block.getWidth.toFloat,
                                     block.getHeight.toFloat)

    if(!cachedItems.contains(z) ||
       (cachedItems(z).size.width != size.width ||
        cachedItems(z).size.height != size.height)) {

      val hints = new TranscodingHints
      hints.put(SVGAbstractTranscoder.KEY_WIDTH, size.width);
      hints.put(SVGAbstractTranscoder.KEY_HEIGHT, size.height);
      hints.put(XMLAbstractTranscoder.KEY_DOM_IMPLEMENTATION,
                SVGDOMImplementation.getDOMImplementation)
      hints.put(XMLAbstractTranscoder.KEY_DOCUMENT_ELEMENT_NAMESPACE_URI,
                SVGConstants.SVG_NAMESPACE_URI)
      hints.put(XMLAbstractTranscoder.KEY_DOCUMENT_ELEMENT,
                SVGConstants.SVG_SVG_TAG)
      hints.put(XMLAbstractTranscoder.KEY_XML_PARSER_VALIDATING,
                false)

      val transcoder = new BufferedImageTranscoder
      transcoder.setTranscodingHints(hints)

      val input = new TranscoderInput(svgScale.svg(z))
      transcoder.transcode(input, null)

      cachedItems += z ->
        new CachedItem(Toolkit.getDefaultToolkit
          .createImage(transcoder.getImage.getSource), size)

    }

    g2.drawImage(cachedItems(z).image,
                 block.getX.toInt, block.getY.toInt,
                 cachedItems(z).image.getWidth(null),
                 cachedItems(z).image.getHeight(null),
                 null)

    // XYBlockRenderer part.
    val entities = state.getEntityCollection
    if(entities != null)
      addEntity(entities, block, dataset, series, item, 0.0, 0.0)

  }

  protected class BufferedImageTranscoder extends ImageTranscoder {

    /** @see [[org.apache.batik.transcoder.image.ImageTranscoder]]. */
    def createImage(width: Int, height: Int) = {

      img = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB)
      img

    }

    /** @see [[org.apache.batik.transcoder.image.ImageTranscoder]]. */
    def writeImage(img: BufferedImage, out: TranscoderOutput) { this.img = img }

    /** SVG as BufferedImage.
     *
     * @return The BufferedImage.
     */
    def getImage = img

    protected var img: BufferedImage = null

  }


  /** @see [[java.awt.geom.Dimension2D]]. */
  protected object Dimension2D {

    /** It's the missed Dimension2D.Float class. */
    protected[XYSVGRenderer] class Float(val width: scala.Float,
                                         val height: scala.Float)

  }

  /** Cache for items. */
  protected class CachedItem(val image: Image, val size: Dimension2D.Float)

  /** Cache items to prevent multiple resizing. */
  protected var cachedItems = Map[Double, CachedItem]()

}

